import os
import tempfile
import slugify
import clisync

from typing import Any, List, Optional, Union, Literal, Tuple

from pyhectiqlab import API_URL
from pyhectiqlab import Config
from pyhectiqlab.project import Project
from pyhectiqlab.artifact import Artifact
from pyhectiqlab.tag import Tag
from pyhectiqlab.model import Model
from pyhectiqlab.dataset import Dataset
from pyhectiqlab.versions import PackageVersion
from pyhectiqlab.step import Step

from pyhectiqlab.client import Client
from pyhectiqlab.decorators import classproperty, functional_alias, no_git_diff
from pyhectiqlab.metrics import MetricsManager
from pyhectiqlab.logging import hectiqlab_logger as logger


class Run:
    _metrics: Optional[MetricsManager] = None
    _tmp_artifacts: Optional[str] = None
    _id: Optional[str] = None
    _rank: Optional[int] = None
    _slug: Optional[str] = None
    _step: Optional[int] = None
    _config: Optional[dict] = None
    _client: Client = Client

    def __init__(
        self,
        rank: Optional[int] = None,
        title: Optional[str] = None,
        project: Optional[str] = None,
        category: Literal[None, "develop", "evaluation", "test", "training"] = None,
        config: Optional[dict] = None,
        reuse: Optional[bool] = False,
    ):
        if rank is not None and isinstance(rank, int):
            self.retrieve(rank, project)
        elif isinstance(rank, str) or rank is None:
            title = title or rank  # support for positional argument
            if title is None:
                logger.warning("No run title or rank provided, using `Untitled run` as title.")
                title = "Untitled run"
            self.create(title=title, project=project, category=category, config=config, reuse=reuse)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._metrics:
            self._metrics.flush_cache()
        if Run.id is None:
            return
        if exc_type is not None:
            if Step.id:
                Step.end(status="failed")
            Run._update(run_id=Run.id, status="failed", package_versions=PackageVersion.all())
        else:
            if Step.id:
                Step.end(status="completed")
            Run._update(run_id=Run.id, status="completed", package_versions=PackageVersion.all())

    @staticmethod
    @functional_alias()
    def track_version(repos: Union[str, List[str], None] = None, run_id: Optional[str] = None):
        """Track the git status of the given repositories.

        Args:
            repos (List[str]): List of modules for which the files are located in a git repertoire.
            run_id (str): Id of the run to update. If None, the current run ID is used. Default: None.
        """
        if not isinstance(repos, list) and (repos is not None):
            repos = [repos]
        if info := PackageVersion.all(repos=repos):
            Run._update(run_id=run_id, package_versions=info)

    @staticmethod
    def setup(**kwargs):
        """Setup the run with the given `id` and `rank`."""
        Run._id = kwargs.get("id")
        Run._rank = kwargs.get("rank")

    @classproperty
    def tmp_artifacts(cls):
        if cls._tmp_artifacts is None:
            cls._tmp_artifacts = tempfile.mkdtemp()
        return cls._tmp_artifacts

    @classproperty
    def rank(cls) -> Optional[int]:
        return cls._rank

    @classproperty
    def slug(cls) -> Optional[str]:
        if cls._slug is None and cls._id is not None:
            run_info = cls.retrieve_by_id(run_id=cls._id)
            cls._slug = slugify.slugify(run_info.get("title")) if run_info is not None else None
        return cls._slug

    @classproperty
    def id(cls) -> Optional[str]:
        return cls._id

    @classproperty
    def category(cls) -> Optional[str]:
        if cls._category is None:
            cls._category = (cls.get(["category"]) or {}).get("category")
        return cls._category

    @classproperty
    def config(cls) -> Optional[dict]:
        return cls._config

    @staticmethod
    @functional_alias("get_run_id")
    def get_id(id: Optional[str] = None):
        return id or Run.id

    @staticmethod
    @functional_alias("get_run_rank")
    def get_rank(rank: Optional[int] = None) -> Optional[int]:
        return rank or Run.rank

    @staticmethod
    @functional_alias("get_run_slug")
    def get_slug(slug: Optional[str] = None) -> Optional[str]:
        return slug or Run.slug

    @staticmethod
    @functional_alias("run_is_ready")
    def ready() -> bool:
        """Check if the run is ready."""
        return Run.id is not None

    @staticmethod
    @functional_alias("create_run")
    @no_git_diff()
    @clisync.include()
    def create(
        title: Optional[str] = None,
        project: Optional[str] = None,
        category: Literal[None, "develop", "evaluation", "test", "training"] = None,
        config: Union[dict, None] = None,
        tags: Optional[List[str]] = None,
        reuse: bool = False,
    ) -> dict:
        """
        Creates a new run.

        Args:
            title (str): The title of the run.
            project (str): The project of the run. If None, the current project is used.
            category (str): The category of the run in [None, 'develop', 'evaluation', 'test', 'training']. Default: None.
            config (dict): The configuration dict of the run. Default: None.
            tags (list): The tags of the run. Default: None.
            reuse (bool): Set to True to reuse the run if the title already exists. Default: False.

        Raises:
            logging.error: If the project is not found, creation is skipped and offline mode is enabled.
        """

        project = Project.get(project)
        if project is None:
            # If project is None, we either are or must be in offline mode.
            if Run._client.online():
                logger.error(f"Could not create run `{title}`, project not found. Continuing in offline mode...")
                Run._client.online(False)
            return
        Project.set(project)

        data = dict(
            title=title,
            project=project,
            category=category,
            config=config or {},
            tags=tags or [],
            get_if_title_collision=reuse,
        )
        run = Run._client.post("/app/runs", json=data, wait_response=True)
        if run is None:
            logger.error("Run failed to create.")
            return False
        if run.get("reused"):
            logger.warning(
                f"`{title}` already exists in this project. You've been connected to this existing run (rank {run.get('rank')})."
            )

        Run.setup(**run)
        Run._update(run_id=Run.id, status="initiated", package_versions=PackageVersion.all())
        return run

    @staticmethod
    @functional_alias("retrieve_run")
    @no_git_diff()
    @clisync.include()
    def retrieve(
        rank: Optional[int] = None,
        project: Optional[str] = None,
        fields: Optional[List[str]] = None,
        setup_with: bool = True,
    ):
        """
        Retrieves a run from the server.
        Use this method to attach to an existing run.

        Args:
            rank (int, optional): The rank of the run. If None, the current run rank is used. Default: None.
            project (str, optional): The project of the run. If None, the project is taken from the context. Default: None.
            fields (List[str]): The fields to retrieve. Default: None.
            setup_with (bool): Set to True to setup the run with the result. Default: True.

        Raises:
            logging.error: If the project is not found, retrieval is skipped and offline mode is enabled.
        """
        rank = rank or Run.rank
        project = Project.get(project)
        if project is None:
            # If project is None, we either are or must be in offline mode.
            if Run._client.online():
                logger.error(f"Could not retrieve run {rank}, project not found. Continuing in offline mode...")
                Run._client.online(False)
            return
        Project.set(project)

        run_info = Run._client.get(
            "/app/runs/retrieve", params=dict(rank=rank, project=project, fields=fields), wait_response=True
        )
        if run_info is None:
            logger.error(f"No run with {rank} found in project {project}")
            return
        if setup_with:
            Run.setup(**run_info)
        return run_info

    @staticmethod
    @functional_alias("retrieve_run_by_id")
    @no_git_diff()
    @clisync.include()
    def retrieve_by_id(
        run_id: Optional[str] = None,
        fields: Optional[List[str]] = None,
        setup_with_result: Optional[bool] = False,
    ):
        """
        Retrieves a run from the server.
        Use this method to attach to an existing run.

        Args:
            id (int): The ID of the run. If None, the current run ID is used. Default: None.
            fields (List[str]): The fields to retrieve. Default: None.
            setup_with_result (bool): Set to True to setup the run with the result. Default: False.

        Raises:
            logging.error: If the run is not found, retrieval is skipped and None is returned.
        """

        run_id = run_id or Run.id
        if run_id is None:
            logger.error("Could not retrieve run by ID, no ID found.")
            return
        run_info = Run._client.get(f"/app/runs/{run_id}", params=dict(fields=fields), wait_response=True)
        if run_info is None:
            logger.error(f"Could not retrieve run {run_id} by ID, run not found.")
            return
        if setup_with_result:
            Run.setup(**run_info)
        return run_info

    @staticmethod
    @functional_alias()
    @clisync.include()
    def retrieve_config_by_id(run_id: Optional[str] = None) -> dict:
        """Retrieve a config from a run.
        If the run does not have a config, an empty config is returned.

        Args:
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): Project name. If None, the current project is used. Default: None.

        Raises:
            logging.error: If the run is not found, returns an empty dict.
        """
        run_info = Run.retrieve_by_id(run=run_id, fields=["config"], setup_with_result=False)
        if run_info is None:
            logger.error(f"Could not retrieve config by run id {run_id}, run not found")
            return {}

        return run_info.get("config", {})

    @staticmethod
    @functional_alias()
    @clisync.include()
    def retrieve_config(rank: Optional[int] = None, project: Optional[str] = None) -> dict:
        """Retrieve a config from a run.
        If the run does not have a config, an empty config is returned.

        Args:
            rank (int): Run rank. If None, the current run rank is used. Default: None.
            project (str, optional): Project name. If None, the current project is used. Default: None.

        Raises:
            logging.error: If the run is not found, returns an empty dict.
        """
        run_info = Run.retrieve(rank=rank, project=project, fields=["config"], setup_with=False)
        if run_info is None:
            logger.error(f"Could not retrieve config from run {rank} in project {project}, run not found.")
            return {}

        return run_info.get("config", {})

    @staticmethod
    @functional_alias("run_exists")
    @clisync.include()
    def exists(rank: int, project: Optional[str] = None) -> bool:
        """
        Check if a run exists.

        Args:
            rank (int): The rank of the run.
            project (str): The project of the run. If None, the project is taken from the context. Default: None.
        """
        res = Run._client.get("/app/runs/retrieve", params=dict(rank=rank, project=project), wait_response=True)
        return res is not None

    @staticmethod
    def _update(run_id: Optional[str] = None, wait_response: bool = False, **kwargs):
        """Update the run.

        Args:
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            wait_response (bool): Set to True to wait for the response. Default: False.
            **kwargs: Additional arguments to update.

        Raises:
            logging.error: If the run is not found, the update is not performed and it returns None.
        """
        run_id = run_id or Run.id
        if run_id is None:
            logger.error("No run to update, ID not found.")
            return
        return Run._client.patch(f"/app/runs/{run_id}", json=kwargs, wait_response=wait_response)

    @staticmethod
    @functional_alias("set_run_category")
    @clisync.include()
    def set_category(category: Literal["develop", "evaluation", "test", "training"], run_id: Optional[str] = None):
        """Set the category of the run."""
        Run._update(run_id=run_id, category=category)

    @staticmethod
    @functional_alias("rename_run")
    @clisync.include()
    def rename(title: str, run_id: Optional[str] = None):
        """Rename the run with a new title.

        Args:
            title (str): The new title of the run.
        """
        Run._update(run_id=run_id, title=title)

    @staticmethod
    @functional_alias()
    @clisync.include()
    def add_artifact(
        path: str,
        name: Optional[str] = None,
        step: Optional[int] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Log a file as an artifacts. If a file already exists with the name, it will be overwritten.

        Args:
            path (str) : Path to the file.
            name (str, optional): Name of the artifact or the group of artifacts. If None, the name is the basename of the file. Default: None.
            step (int, optional): The optional step stamp of the artifacts. If None, the artifacts is not considered as a step artifact. Default: None.
            run_id (str, optional): ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the artifact. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to upload sync. If False, the upload is made in background. Default: False.
            verbose (bool): Set to True to show the upload progress. Default: True.

        Raises:
            logging.error: If the project is not found or if the run ID is not found, the artifact creation is not performed.
        """
        run_id = Run.get_id(run_id)
        project = Project.get(project)
        if project is None:
            logger.error("Could not create artifact, project not found.")
            return
        if run_id is None:
            logger.error("Could not create artifact, run ID not found.")
            return
        return Artifact.create(
            path,
            Run.get_id(run_id),
            Project.get(project),
            name=name,
            group=name,
            step=step,
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias()
    def add_figure(
        figure,
        name: str,
        step: Optional[int] = None,
        dpi: int = 200,
        extension: str = "png",
        wait_response: bool = False,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        **kwargs,
    ):
        """
        Save a matplotlib figure on a temp dir and push the image to the lab.

        Args:
            figure (matplotlib.pyplot.Figure): The figure to save.
            name (str): The name of the artifact.
            step (int): The optional step stamp of the artifacts. If None, the artifacts is not considered as a step artifact. Default: None.
            dpi (int): The dpi of the figure. Default: 200.
            extension (str): The extension of the figure. Default: "png".
            wait_response [bool]: Set to true to upload sync. If False, the upload is made in background. Default: False.
            run_id (str): The ID of the run. If None, the current run ID is used. Default: None.
            project (str): The project of the artifact. If None, the current project is used. Default: None.
            **kwargs: Additional arguments to pass to the savefig method.
        """

        basename, name_ext = os.path.splitext(name)
        if name_ext != "":
            extension = name_ext.strip(".")
        else:
            extension = (extension or "png").strip(".")

        slug_name = slugify.slugify(basename)
        fname = os.path.join(Run.tmp_artifacts, slug_name + "." + extension)
        figure.savefig(fname, dpi=dpi, format=extension, bbox_inches="tight", **kwargs)

        return Run.add_artifact(
            fname, name=name, step=step, wait_response=wait_response, run_id=run_id, project=project
        )

    @staticmethod
    @functional_alias()
    def add_config(config: Union[Config, dict[str, Any]], run_id: Optional[str] = None):
        """Attach a config to the run."""
        if isinstance(config, Config):
            config = config.to_dict()
        Run._config = config
        return Run._update(run_id=run_id, config=config)

    @staticmethod
    @functional_alias()
    def add_metric(key: str, step: int, value: float):
        """
        Add a metric to the run.

        Args:
            key (str): The key of the metric.
            step (int): The step of the metric.
            value (float): The value of the metric.

        Note:
            The metrics are not pushed immediately. They are pushed when the `Run._push_metrics`
            method is called inside the MetricsManager. See `pyhectiqlab.metrics.MetricsManager`.
        """
        if Run._metrics is None:
            Run._metrics = MetricsManager(push_method=Run._push_metrics, run_id=Run.id)
        return Run._metrics.add(key, step, value)

    @staticmethod
    @functional_alias()
    def get_agg_metrics(key: str, run_id: Optional[str] = None) -> List[Tuple[Union[int, float], Union[int, float]]]:
        """Get the aggregated metrics for the given key.

        Args:
            key (str): The key of the metric.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.

        Returns:
            list[tuple[number]]: The aggregated metrics.

        Raises:
            logging.error: If the run ID is not found, it returns an empty list.
        """
        run_id = run_id or Run.id
        if run_id is None:
            logger.error("No run to get aggregated metrics from.")
            return []
        # This does not work. Need to fix it.
        return MetricsManager.get_agg(key, run_id)

    @staticmethod
    def _push_metrics(
        key: str,
        values: List[Tuple[Union[int, float], Union[int, float], Union[int, float]]],
        run_id: Optional[str] = None,
    ):
        """Push the metrics to the server.
        Do not use this method directly. Use the `add_metric` method instead.
        values is a list of tuples (step, value, timestamp).

        Args:
            key (str): The key of the metric.
            values (List[Tuple[Union[int, float], Union[int, float], Union[int, float]]): The values of the metric.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
        """
        run_id = run_id or Run.id
        if key is None or run_id is None or len(values) == 0:
            return
        data = dict(name=key, values=values, run=run_id)
        Run._client.post("/app/metrics", json=data, wait_response=False)

    @staticmethod
    @functional_alias("start_step")
    def start(
        name: str,
        run_id: Optional[str] = None,
        description: Optional[str] = None,
        metadata: Optional[dict] = None,
        status: Optional[str] = None,
    ):
        """
        Start a new step.

        Args:
            title (str): Title of the step.
        """
        if Step.id is not None:
            Step.end()

        return Step.start(
            name=name,
            run_id=Run.get_id(run_id),
            description=description,
            metadata=metadata,
            status=status,
        )

    @staticmethod
    @functional_alias("end_step")
    def end(run_id: Optional[str] = None, status: Optional[str] = None):
        """End the current step."""
        return Step.end(status=status)

    @staticmethod
    @functional_alias()
    def attach_model(
        name: str,
        version: str,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Attach a model to the run.

        Args:
            name (str): The name of the model.
            version (str): The version of the model. If None, the latest version is attached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Model.attach(
            name=name,
            version=version,
            run_id=Run.get_id(run_id),
            project=Project.get(project),
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias()
    def detach_model(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Detach a model from the run.

        Args:
            name (str): The name of the model.
            version (str): The version of the model. If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """

        return Model.detach(
            name=name,
            version=version,
            run_id=Run.get_id(run_id),
            project=Project.get(project),
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias()
    def attach_dataset(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Attach a dataset to the run.

        Args:
            name (str): The name of the dataset.
            version (str): The version of the dataset.  If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Dataset.attach(
            name=name,
            version=version,
            run_id=Run.get_id(run_id),
            project=Project.get(project),
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias()
    def detach_dataset(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Detach a dataset to the run.

        Args:
            name (str): The name of the dataset.
            version (str): The version of the dataset.  If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Dataset.detach(
            name=name,
            version=version,
            run_id=Run.get_id(run_id),
            project=Project.get(project),
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias("set_run_status")
    @clisync.include()
    def set_status(status: str, run_id: Optional[str] = None):
        """Set the run to the given status."""
        Run._update(run_id=run_id, status=status)

    @staticmethod
    @functional_alias("failed")
    @clisync.include()
    def failed(run_id: Optional[str] = None):
        """Set the run to the failed status."""
        Run.set_status(run_id=run_id, status="failed")

    @staticmethod
    @functional_alias("stopped")
    @clisync.include()
    def stopped(run_id: Optional[str] = None):
        """Set the run to the stopped status."""
        Run.set_status(run_id=run_id, status="stopped")

    @staticmethod
    @functional_alias("completed")
    @clisync.include()
    def completed(run_id: Optional[str] = None):
        """Set the run to the completed with success status."""
        Run.set_status(run_id=run_id, status="completed")

    @staticmethod
    @functional_alias("pending")
    @clisync.include()
    def pending(run_id: Optional[str] = None):
        """Set the run to the pending status."""
        Run.set_status(run_id=run_id, status="pending")

    @staticmethod
    @functional_alias("running")
    @clisync.include()
    def running(run_id: Optional[str] = None):
        """Set the run to the running status."""
        Run.set_status(run_id=run_id, status="running")

    @staticmethod
    @functional_alias("training")
    @clisync.include()
    def training(run_id: Optional[str] = None):
        """Set the run to the training status."""
        Run.set_status(run_id=run_id, status="training")

    @staticmethod
    @functional_alias("add_tags_to_run")
    @clisync.include()
    def add_tags(tags: Union[str, List[str]], run_id: Optional[str] = None, project: Optional[str] = None):
        """Add a tag to the run.
        For functional alias, use `attach_tags_to_run`.

        Args:
            title (str): The new title of the run.
            run_id (str): The ID of the run. If None, the current run ID is used. Default: None.
            project (str): The project of the run. If None, the current project is used. Default: None.

        Raises:
            logging.error: If the run ID is not found, the tag is not attached.
        """
        if isinstance(tags, str):
            tags = [tags]

        run_id = Run.get_id(run_id)
        if run_id is None:
            logger.error("No run to attach tags to.")
            return
        Tag.attach_to_run(tags, run_id=Run.get_id(run_id), project=Project.get(project))

    @staticmethod
    @functional_alias("detach_tag_from_run")
    @clisync.include()
    def detach_tag(tag: str, run_id: Optional[str] = None, project: Optional[str] = None):
        """Remove a tag from the run.
        For functional alias, use `detach_tag_from_run`.

        Args:
            title (str): The new title of the run.
            run_id (str): The ID of the run. If None, the current run ID is used. Default: None.
            project (str): The project of the run. If None, the current project is used. Default: None.

        Raises:
            logging.error: If the run ID is not found, the tag is not detached.
        """

        run_id = Run.get_id(run_id)
        if run_id is None:
            logger.error("No run to attach tags to.")
            return
        Tag.detach_from_run(tag=tag, run_id=run_id, project=Project.get(project))

    def __str__(self):
        pad = 8
        path = "/" + Project.get() + "/runs/" + str(self.rank)  # TBD
        return (
            f"<Run {self.rank}>"
            f"\n{'Project'.ljust(pad)}: {Project.get()}"
            f"\n{'author'.ljust(pad)}: {self._client.auth.user}"
            f"\n{'online'.ljust(pad)}: {self._client.auth.online}"
            f"\n{'url'.ljust(pad)}: {API_URL+path} "
        )



import os
import socket
import asyncio
import logging
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, FileSizeColumn, TimeRemainingColumn
from typing import List, Optional

from pyhectiqlab.settings import getenv
from pyhectiqlab.decorators import classproperty

logger = logging.getLogger()


def is_running_event_loop() -> bool:
    """Check if an event loop is running."""
    try:
        return asyncio.get_event_loop().is_running()
    except RuntimeError:
        return False


def batched(iterable, n):
    """Yield successive n-sized batches from iterable."""
    for i in range(0, len(iterable), n):
        i1 = min(i + n, len(iterable))
        yield iterable[i:i1]


def list_all_files_in_dir(local_folder: str) -> List[str]:
    """List all files in a directory.
    If the path is a file, the file is returned.

    Args:
        local_folder (str): The folder to list files from.
    """
    if not os.path.exists(local_folder):
        raise FileNotFoundError(f"Directory {local_folder} does not exist.")
    filenames = []
    if not os.path.isdir(local_folder):
        # If the path is a file, return the file
        return [local_folder]

    for el in os.walk(local_folder):
        for f in os.listdir(el[0]):
            complete_path = os.path.join(el[0], f)
            if not os.path.isdir(complete_path):
                if os.path.isfile(complete_path):
                    filenames.append(complete_path)
    return filenames


def extract_host_from_source(source: str) -> str:
    """Extract the resource from a path.

    If the path is a cloud path (e.g., s3://bucket/dataset), the resource is inferred from the path.
    If the path is a local path, the resource is set to "local" if not provided.

    Args:
        path (str): The path to extract the resource from.
        default_resource (str): The resource to compare the path with.
    """
    if source.startswith("s3://"):
        return "s3"
    elif source.startswith("gs://"):
        return "gs"
    else:
        return socket.gethostname()


async def queue_gather(tasks: List[callable], workers: Optional[int] = 10):
    """Gather tasks using a queue and workers.
    The tasks will be placed in a queue and workers will be
    created to process the tasks.
    The workers will be stopped when the queue is empty.

    Example usage:

    ```python
    async def task():
        await asyncio.sleep(1)
        return "done"
    tasks = [task() for _ in range(10)]
    results = await queue_gather(tasks)
    print(results)
    ```

    Args:
        tasks (List[callable]): A list of async methods
        workers (Optional[int], optional): The number of workers. Defaults to 10.
    """

    async def worker(i: int, queue: asyncio.Queue):
        """Worker to process tasks from a queue."""
        results = []
        while not queue.empty():
            logger.debug(f"Worker {i}. Queue size: {queue.qsize()}")
            idx, task = await queue.get()
            try:
                result = await task
                results.append((idx, result))
            except Exception as e:
                print(e)
                results.append((idx, None))
            finally:
                queue.task_done()
        return results

    queue = asyncio.Queue()
    [queue.put_nowait((i, task)) for i, task in enumerate(tasks)]
    # Create workers
    workers = [asyncio.create_task(worker(idx, queue)) for idx in range(workers)]
    await queue.join()
    for worker in workers:
        worker.cancel()
    results = await asyncio.gather(*workers, return_exceptions=True)
    flatten_results = [result for worker_results in results for result in worker_results]
    flatten_results.sort(key=lambda x: x[0])
    return [result for (_, result) in flatten_results]


class ProgressHandler:
    """Progress handler. This class wraps the rich Progress class and provides
    an interface to control the progress bar within `pyhectiqlab`.

    The progress bar can be hidden by setting the `hide` parameter to `True`.

    Note:
        When used jointly with `register_progress`, the progress object is not owned
        by the `ProgressHandler` object and it is not the `ProgressHandler`
        responsibility to start or stop the external progress. Therefore, the wrapped
        `Progress` object will not be started when calling `start` and not be stopped
        when calling `stop`.


    Example usage:
        ```python
        progress = ProgressHandler(hide=True)
        task_id = progress.add_task("Downloading...", total=10)
        for i in range(10):
            progress.update(task_id, advance=1)
        progress.stop_task(task_id)
        progress.stop()
        ```

    """

    _external_progress: Optional[Progress] = None

    def __init__(self):
        self.progress = None if getenv("HECTIQLAB_HIDE_PROGRESS") else self.prepare_progress()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @classmethod
    def using_external(cls):
        return cls._external_progress is not None

    @classproperty
    def external_progress(cls):
        return cls._external_progress

    @staticmethod
    def prepare_progress():
        """Prepares the progress object. If the external progress is used, it returns it. Otherwise
        it returns a new progress object."""
        # priority to external progress
        if ProgressHandler.using_external():
            return ProgressHandler.external_progress
        progress = Progress(
            TextColumn("[bold bright_blue]{task.description}", justify="left"),
            BarColumn(),
            TaskProgressColumn(text_format="[progress.iteration]{task.percentage:>3.0f}%", show_speed=True),
            FileSizeColumn(),
            TimeRemainingColumn(compact=False, elapsed_when_finished=True),
            transient=False,
        )
        return progress

    def add_task(self, desc: str, total: int):
        """Adds a task with `desc` and `total` to the progress bar. See `rich.progress.Progress.add_task`
        for more information."""
        if self.progress is None:
            return
        return self.progress.add_task(desc, total=total)

    def update(self, task_id: Optional[int], advance: int):
        """Updates the progress of a task with `task_id` by `advance`. See `rich.progress.Progress.update`
        for more information."""
        if self.progress is None or task_id is None:
            return
        self.progress.update(task_id, advance=advance)

    def stop_task(self, task_id: Optional[int]):
        """Stops the task with `task_id`. See `rich.progress.Progress.stop_task` for more information."""

        if self.progress is None or task_id is None:
            return
        self.progress.stop_task(task_id)
        if self.progress.live.transient or self.using_external():
            self.progress.remove_task(task_id)

    def start(self):
        """Starts the progress bar."""
        if self.progress is None or self.using_external():
            return
        self.progress.start()

    def stop(self):
        """Stops the progress bar."""
        # Cannot stop progress if is None or it is used elsewhere
        if self.progress is None or self.using_external():
            return
        self.progress.stop()


def register_progress(progress: Progress):
    """Register a `rich.progress.Progress` object."""
    if isinstance(progress, Progress):
        ProgressHandler._external_progress = progress



import slugify
import clisync
import logging

from typing import Optional, List
from pyhectiqlab.settings import getenv
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import classproperty, functional_alias
from pyhectiqlab.logging import hectiqlab_logger as logger


class Project:
    """
    Project class for handling project related operations.

    Attributes:
        slug (str): Slug of the project.
    """

    _slug: str = None
    _repos: Optional[List[str]] = None
    _client: Client = Client

    @classproperty
    def slug(cls) -> str:
        if cls._slug is None and getenv("HECTIQLAB_PROJECT") is not None:
            cls.set(getenv("HECTIQLAB_PROJECT"))
        return cls._slug

    @classmethod
    def allow_dirty(cls) -> bool:
        if not Client.online():
            return True
        if getenv("HECTIQLAB_ALLOW_DIRTY") == "no":
            return False
        return True

    @classmethod
    def repos(cls) -> List[str]:
        if cls._repos is None and getenv("HECTIQLAB_REPOS") is not None:
            cls._repos = getenv("HECTIQLAB_REPOS").split(",")
        return cls._repos

    @staticmethod
    @clisync.include()
    def create(
        name: str,
        repo: Optional[str] = None,
        force_no_git_diff: bool = True,
    ) -> Optional[str]:
        """
        Create a project.

        Args:
            name (str): Name of the project.
            repo (str, optional): Repository URL. Default: None.
            force_no_git_diff (bool): Force no git diff. Default: True.
        """
        from pyhectiqlab.const import DISABLE_PROJECT_CREATION

        if DISABLE_PROJECT_CREATION:
            logger.error("Project creation is disabled. Continuing anyway...")
            return
        slug = "/".join([slugify.slugify(n) for n in name.split("/")])
        body = {
            "slug": slug,
            "name": name,
            "repo": repo,
            "force_no_git_diff": force_no_git_diff,
        }
        project = Project._client.post("/app/projects", json=body, wait_response=True)
        if project is None:
            logger.error("Failed to create project.")
            return
        logger.info(f"Project `{slug}` created.")
        return slug

    @staticmethod
    def exists(slug: str) -> bool:
        project = Project._client.get(f"/app/projects/{slug}", wait_response=True)
        if project is None:
            return False
        return True

    @staticmethod
    @functional_alias("retrieve_project")
    @clisync.include()
    def retrieve(
        slug: str,
        fields: Optional[List[str]] = [],
    ) -> Optional[dict]:
        """Retrieve a project info.

        Args:
            slug (str): Slug of the project.
            fields (List): Fields to retrieve.
        """
        info = Project._client.get(f"/app/projects/{slug}", params={"fields": fields}, wait_response=True)
        return info

    @staticmethod
    @functional_alias("get_project")
    def get(slug: Optional[str] = None):
        return slug or Project.slug

    @staticmethod
    @functional_alias("set_project")
    def set(slug: str):
        if slug is None:
            return
        if not Project.exists(slug):
            # If project does not exist, we either are or must be offline.
            if Project._client.online():
                logger.error(f"The project `{slug}` does not exist. Continuing in offline mode.")
                Project._client.online(False)
            return
        Project._slug = slug



import clisync

from typing import Optional, List, Union
from functools import partial

from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, classproperty
from pyhectiqlab.logging import stream_log, get_handler

import logging

logger = logging.getLogger()


class Step:
    _id: Optional[str] = None
    _is_listening_logs: bool = False
    _client: Client = Client

    @classproperty
    def id(cls):
        return cls._id

    def __init__(
        self,
        name: Optional[str] = "Untitled step",
        description: Optional[str] = None,
        metadata: Optional[dict] = None,
        status: Optional[str] = None,
        run_id: Optional[str] = None,
    ):
        self.start(name=name, description=description, metadata=metadata, status=status, run_id=run_id)

    @staticmethod
    @functional_alias("start_step")
    def start(
        name: Optional[str] = "Untitled step",
        run_id: Optional[str] = None,
        description: Optional[str] = None,
        metadata: Optional[dict] = None,
        status: Optional[str] = None,
    ) -> int:
        """
        Start a new step.

        Args:
            title (str): Title of the step.

        Raises:
            logger.error: If no run ID is found, it does not create the step and returns None.
        """
        from pyhectiqlab.run import Run

        run_id = Run.get_id(run_id)
        if run_id is None:
            logger.error("No active run found. Skipping step creation.")
            return

        if Step.id is not None:
            Step.end()

        json = {
            "name": name,
            "description": description,
            "metadata": metadata,
            "status": status,
            "run": run_id,
        }
        step = Step._client.post("/app/steps", wait_response=True, json=json)
        Step.id = step["id"]

        if Step._is_listening_logs:
            # Stop the previous listener
            handler = get_handler()
            if handler is not None:
                handler.close()
            Step._is_listening_logs = False

        # Start listening to the logs
        stream_log(on_dump=partial(Step._handle_log_dump, step=Step.id))
        Step._is_listening_logs = True
        return step

    @staticmethod
    @functional_alias("end_step")
    def end(status: Optional[str] = None, step: Optional[str] = None):
        """
        End the current step.

        Args:
            step (str, optional): Step name. If None, the current step is used. Default: None.

        Raises:
            logger.error: If no step ID is found, it does nothing and returns None.
        """
        step = step or Step.id
        if step is None:
            if Step._client.online():
                logger.error("No active step found. Skipping step end.")
            return

        status = status or "completed"
        handler = get_handler()
        logs = handler.value() if handler is not None else None
        if logs is not None and len(logs) > 0:
            if isinstance(logs, list):
                logs = "\n".join(logs)

        Step._client.post(f"/app/steps/{step}/end", wait_response=True, json={"status": status, "logs": logs})
        Step._id = None
        if handler:
            handler.close(flush=False)
            Step._is_listening_logs = False
        return

    @staticmethod
    @functional_alias("update_step")
    @clisync.include()
    def update(
        name: Optional[str] = None,
        description: Optional[str] = None,
        logs: Optional[str] = None,
        status: Optional[str] = None,
        metadata: Optional[dict] = None,
        step: Optional[str] = None,
    ):
        """
        Update the logs of the current step.

        Args:
            name (str, optional): Name of the step. If None, the current step is used. Default: None.
            description (str, optional): Description of the step. Default: None.
            logs (Union[str, List[str]]): Logs to update.
            status (str, optional): Status of the step. Default: None.
            metadata (dict, optional): Metadata to update. Default: None.
            step (str, optional): Step name. If None, the current step is used. Default: None.

        Raises:
            logger.error: If no active step is found, it does nothing and returns None.
        """
        step = step or Step.id
        if step is None:
            logger.error("No active step found. Skipping step update.")
            return
        body = {"name": name, "description": description, "logs": logs, "status": status, "metadata": metadata}
        body = {k: v for k, v in body.items() if v is not None}
        if len(body) == 0:
            return
        Client.put(f"/app/steps/{step}", wait_response=True, json=body)

    @staticmethod
    def _handle_log_dump(logs: Union[str, List[str]], step: Optional[str] = None):
        # Format the logs
        if len(logs) == 0:
            return
        if isinstance(logs, list):
            logs = "\n".join(logs)
        Step.update(logs=logs, step=step)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.end(status="failed")
        else:
            self.end(status="completed")



from .run import *
from .artifact import *
from .model import *
from .dataset import *
from .project import *
from .step import *
from .tag import *
from .message import *

from pyhectiqlab.utils import register_progress



from .utils import expose_static_methods
from pyhectiqlab.message import Message

"""
Functional message API. To expose a function, use the `exposed_method` decorator in the singleton.
"""

locals().update(expose_static_methods(Message))



import pytest
import os
import time
import logging
logger = logging.getLogger()

@pytest.fixture
def path():
    return os.path.abspath(os.path.join(__file__, "../dummy/artifact.txt"))


def test_messages():

    from pyhectiqlab import functional as hl, Run

    run = Run("Test receiving/publishing messages", project="hectiq-ai/test")
    logger.warning(f"In your bash, type: `HECTIQLAB_API_URL={os.getenv('HECTIQLAB_API_URL')} hectiq-lab Message.publish --run {run.id} --key 'test' --value 'Hello world!'`")
    logger.warning("Waiting for message...")
    msg = None
    while not msg:
        msg = hl.get_message(key="test")
        time.sleep(1)
    if msg:
        logger.warning(f"Received message: {msg}")
        msg.ack()

if __name__ == "__main__":
    test_messages()


from collections import defaultdict
from typing import Callable, List, Literal, Optional, Tuple, Union
from functools import partial

import threading
import numpy as np
import time
import datetime as dt
import logging

logger = logging.getLogger("pyhectiqlab")


class RepeatedTimer(threading.Thread):
    """
    RepeatedTimer class is used to run a method at a given interval.

    Args:
        method (callable): Method to run.
        delay (int): Delay between two executions.
    """

    def __init__(self, method: Callable[[], None], delay: int):
        super().__init__()
        self.daemon = True
        self._stopper = threading.Event()
        self.delay = delay
        self.method = method

    def run(self):
        while not self._stopper.wait(self.delay):
            self.method()

    def stopped(self):
        return self._stopper.is_set()

    def stop(self):
        self._stopper.set()


class MetricsManager:
    """
    MetricsManager class is used to manage the metrics cache and push data to the server.

    Args:
        push_method (callable, optional): Function to push data to the server
        max_cache_timeout (int): Maximum number of seconds in the cache
        max_cache_length (int): Number of elements in the cache to trigger a cache dump
        min_cache_flush_delay (int): Minimum number of seconds between two dumps
        aggregate_metrics (str): Wheter to push all data in cache or aggregate
    """

    __max_cache_quantity_per_second = 200

    def __init__(
        self,
        push_method: Callable[[str, List[float]], None],
        run_id: Optional[str] = None,
        max_cache_timeout: int = 10,
        max_cache_length: int = 100,
        min_cache_flush_delay: int = 5,
        aggregate_metrics: str = "none",
    ):
        self.cache = defaultdict(list)
        self.cache_last_push = {}
        self._warning_is_shown = False
        self.max_cache_length = max_cache_length
        self.min_cache_flush_delay = min_cache_flush_delay

        self.push_method = partial(push_method, run_id=run_id)
        self.set_aggr(aggregate_metrics)

        self.timer = RepeatedTimer(self.flush_cache, max_cache_timeout)
        self.timer.start()
        return

    def update_cache_settings(
        self,
        max_cache_timeout: Optional[int] = None,
        max_cache_length: Optional[int] = None,
        min_cache_flush_delay: Optional[int] = None,
    ):
        """
        Update cache settings.
        """

        if max_cache_timeout is not None:
            self.max_cache_timeout = max_cache_timeout
            self.timer.stop()
            self.timer = RepeatedTimer(self.flush_cache, max_cache_timeout)
            self.timer.start()
        if max_cache_length is not None:
            self.max_cache_length = max_cache_length
        if min_cache_flush_delay is not None:
            self.min_cache_flush_delay = min_cache_flush_delay

    def set_aggr(self, new_value: Literal["none", "sum", "max", "mean"]):
        vals = ["none", "sum", "max", "mean"]
        assert new_value in vals, f"Aggr must be in {vals}"
        self.aggregate_metrics = new_value

    def get_agg(self, key: str, run_id: str) -> List[Tuple[Union[int, float], Union[int, float]]]:
        # raise NotImplementedError()
        return []

    def add(self, key: str, step: int, value: float) -> None:
        """
        Add a new metric to the cache.
        """
        if value is None:
            return
        self.cache[key].append((float(step), float(value), dt.datetime.now(tz=dt.timezone.utc).timestamp()))
        if len(self.cache[key]) > self.max_cache_length:
            if key in self.cache_last_push:
                last_push_time = self.cache_last_push[key]
                elapsed = time.time() - last_push_time
                if elapsed < self.min_cache_flush_delay:
                    return
            self.push_data(key)

    def subsample(self, data: List[float]) -> List[float]:
        normalized_max = int(self.__max_cache_quantity_per_second * self.min_cache_flush_delay)
        if len(data) <= normalized_max:
            return data
        if self._warning_is_shown == False:
            logger.warning(
                f"Your metrics rate ({len(data)//self.min_cache_flush_delay} metrics/second) is exceeding the "
                "maximum rate ({MetricsManager.__max_cache_quantity_per_second}/second). Your metrics have been "
                "subsampled automatically."
            )
            logger.warning(
                "Consider reducing the number of metrics you push or setting an aggregation method with "
                "`run.set_metrics_aggr('mean')`."
            )
            logger.warning("See the documentation for details https://docs.hectiq.ai/objects/metrics/.")
            self._warning_is_shown = True
        indexes = np.linspace(0, len(data) - 1, normalized_max).astype(int)
        return np.array(data)[indexes].tolist()

    def push_data(self, key: str) -> None:
        self.cache_last_push[key] = time.time()
        if self.push_method is None:
            self.cache[key] = []
            if self._warning_is_shown == False:
                logger.warning(
                    "No push method has been set. Please set a push method with `run.set_metrics_push_method`."
                )
            return

        if self.aggregate_metrics == "none":
            data = self.subsample(self.cache[key])
            self.push_method(key, data)
        elif self.aggregate_metrics == "mean":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.mean(s)), float(np.mean(s)))])
        elif self.aggregate_metrics == "max":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.max(s)), float(np.mean(s)))])
        elif self.aggregate_metrics == "sum":
            k = [_[0] for _ in self.cache[key]]
            s = [_[1] for _ in self.cache[key]]
            t = [_[2] for _ in self.cache[key]]
            if len(k) > 0:
                self.push_method(key, [(float(np.max(k)), float(np.sum(s)), float(np.mean(t)))])
        self.cache[key] = []

    def __delete__(self):
        self.timer.stop()
        self.flush_cache()

    def flush_cache(self):
        for key in self.cache:
            self.push_data(key)



import pytest
from utils import random_uuid
from mock_client import mock_client


@pytest.fixture
def slug():
    return "hectiq-ai/test2"


@pytest.fixture
def teardown_config():
    yield
    import os

    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "dummy/configs.toml"))
    if os.path.exists(path):
        os.remove(path)


def test_project_create():
    from pyhectiqlab.const import DISABLE_PROJECT_CREATION
    from pyhectiqlab.project import Project
    from pyhectiqlab.client import Client
    import uuid

    id = str(uuid.uuid4())[:6]
    if DISABLE_PROJECT_CREATION:
        return

    with mock_client(Client):
        project = Project.create(name="test" + id)
        assert project["name"] == "test" + id
        assert project["slug"] == "test" + id


def test_project_retrieve(slug):
    from pyhectiqlab.project import Project

    project = Project.retrieve(slug)
    assert project["name"] == slug
    assert project["slug"] == slug


def test_project_exists(slug):
    from pyhectiqlab.project import Project
    import uuid

    fake_id = str(uuid.uuid4())[:6]

    assert Project.exists(slug)
    assert not Project.exists(slug + fake_id)


def test_project_set_get(slug):
    from pyhectiqlab.project import Project

    Project._slug = None

    assert Project.get() == None
    Project.set(slug)
    assert Project.get() == slug


if __name__ == "__main__":
    pass



from typing import Any, Dict, Callable


def expose_static_methods(cls) -> Dict[str, Callable[..., Any]]:
    """Expose all static methods of a class to a module namespace."""
    namespace = {}
    for name, method in cls.__dict__.items():
        if method.__class__ != staticmethod:
            continue
        method = getattr(cls, name)
        if not hasattr(method, "exposed") or not method.exposed:
            continue
        namespace[method.alias or method.__name__] = method
    return namespace



from functools import wraps
from typing import Optional, Union

from pyhectiqlab.settings import getenv
from pyhectiqlab.logging import hectiqlab_logger


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


def execute_online_only(f):
    """Decorator to execute a function only if the client is online.
    If the client is offline, the function is not executed and a log message is displayed.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        from pyhectiqlab.client import Client

        if not Client.online():
            if not hasattr(f, "_show_online_only_error"):
                if getenv("HECTIQLAB_LOG_LEVEL", "warning").lower() == "debug":
                    msg = f"`{f.__name__}` with args {list(args)} and kwargs {kwargs} not executed. Offline mode."
                else:
                    msg = f"`{f.__name__}` not executed. Offline mode."
                hectiqlab_logger.error(msg)
                f._show_online_only_error = False
            return
        return f(*args, **kwargs)

    return wrapper


def no_git_diff(force: bool = False):

    def _wrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            from pyhectiqlab.project import Project
            from pyhectiqlab.client import Client
            from pyhectiqlab.versions import PackageVersion

            if force:
                return f(*args, **kwargs)
            if not Project.allow_dirty() and PackageVersion.is_dirty():
                hectiqlab_logger.error(f"`{f.__name__}` not executed. There are uncommited changes in the repository.")
                Client.online(False)
            return f(*args, **kwargs)

        return wrapper

    return _wrapper


def beta(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        hectiqlab_logger.error("This method is still in development and can be unstable.")
        return f(self, *args, **kwargs)

    return wrapper


def will_be_depreciated(suggested_method):
    def _method(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            hectiqlab_logger.error(
                f"The method `{f.__name__}` will be removed in a future release. Use `{suggested_method}` instead."
            )
            return f(self, *args, **kwargs)

        return wrapper

    return _method


def functional_alias(alias: Optional[str] = None):
    def _exposed_method(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        wrapper.exposed = True
        wrapper.alias = alias or f.__name__
        return wrapper

    return _exposed_method


def request_handle(f: callable):
    """Use this over httpx.request to handle the response and the errors.
    Instead of returning the response, return the response.json() if the status code is 200.
    If the status code is not 200, log the error and return None.
    """

    def wrapper(*args, **kwargs) -> Union[dict, bytes, None]:
        response = f(*args, **kwargs)
        if response.status_code != 200:
            hectiqlab_logger.error(f"Request failed with status code {response.status_code}: {response.content}")
            return None
        try:
            return response.json()
        except:
            return response.content

    return wrapper



import click


@click.group()
def auth_group():
    pass


@auth_group.command("authenticate", help="Authenticate to the Hectiq Lab.")
@click.option(
    "--generate",
    "only_generate",
    is_flag=True,
    default=False,
    help="If True, the key is only generated and not saved on your computer. Use this option if you want to use the key on another device.",
)
def authenticate(only_generate: bool):
    """Authenticate to the Hectiq Console."""
    import httpx
    import os
    import toml
    from pathlib import Path
    from pyhectiqlab.auth import is_authenticated
    from pyhectiqlab import API_URL
    from pyhectiqlab.settings import getenv

    if only_generate:
        click.secho(
            "🚨 Mode is set to generate only. The key will not be saved on your computer. At the end of the authentification, the key will be shown.",
            fg="cyan",
        )
    else:
        click.secho("👋 Welcome!")

    is_logged = is_authenticated()

    if is_logged and not only_generate:
        # Ask if the user wants to add a new key
        click.secho("You are already logged in.", fg="green")
        should_continue = click.prompt(
            "Do you still want to continue and create a new API key?",
            default="y",
            show_default=True,
            type=click.Choice(["y", "n"]),
        )
        if should_continue == "n":
            return

    # Accessing the user using basic authentication
    click.secho("Please enter your credentials.")
    email = click.prompt("Email", type=str)
    password = click.prompt("Password", type=str, hide_input=True)

    try:
        import socket

        name = socket.gethostname()
    except:
        name = "[unknown hostname]"
    credentials_path = getenv("HECTIQLAB_CREDENTIALS", os.path.join(Path.home(), ".hectiq-lab", "credentials.toml"))
    os.makedirs(os.path.dirname(credentials_path), exist_ok=True)

    # Get the API key
    auth = httpx.BasicAuth(username=email, password=password)
    body = dict(name=name)
    res = httpx.post(f"{API_URL}/app/auth/api-keys", json=body, auth=auth)
    if res.status_code != 200:
        click.echo("Authentication failed.")
        click.echo(res)
        return
    api_key = res.json()

    # Save the key in .hectiq-lab/credentials
    if not only_generate:
        with open(credentials_path, "a") as f:
            # Dump as TOML
            data = {}
            data[name] = api_key
            toml.dump(data, f)
            f.write("\n")
            click.echo(f"A new API key has been added to {credentials_path}.")
    else:
        click.secho("--------------------------------------------------------------------\n🔑 Your API key:")
        click.secho(f"{api_key['value']}", bold=True)
        click.secho("--------------------------------------------------------------------\n")
        click.secho(
            "⚠️ The value above is your API key. It won't be show again and the key has not been saved on your computer.",
            bg="red",
            fg="white",
            bold=True,
        )
        click.echo(
            "Use the env `HECTIQLAB_API_KEY` to authenticate with this key or visit https://docs.hectiq.ai/authenticate.html to learn how to use the key."
        )
        return

    click.secho("You are now logged in.", fg="green")
